// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:nil -*-

#include <cstdlib>

#include <fstream>
#include <math.h>
#include <vector>
#include <visp3/core/vpHomogeneousMatrix.h>
#include <visp3/core/vpException.h>

#include <yarp/math/Math.h>
#include <yarp/os/ResourceFinder.h>
#include <yarp/os/RFModule.h>

#include <yarp/sig/Matrix.h>

#include <ForwardKinematics.h>

#include <yarp/os/Network.h>
#include <yarp/os/LogStream.h>
#include <yarp/os/RFModule.h>
#include <yarp/os/Bottle.h>
#include <yarp/os/Stamp.h>
#include <yarp/sig/Image.h>
#include <yarp/os/PeriodicThread.h>
#include <yarp/os/BufferedPort.h>
#include <yarp/sig/Vector.h>
// #include <yarp/cv/Cv.h>

// #include <opencv2/core/core.hpp>
// #include <opencv2/core/mat.hpp>
// #include <opencv2/opencv.hpp>

using namespace std;
using namespace yarp::os;
using namespace yarp::sig;
using namespace yarp::math;
using namespace yarp::dev;
using namespace iCub::ctrl;

float median(vector<float> vec){

    int size = vec.size();
    sort(vec.begin(), vec.end());
    int mid = size / 2;
    return size % 2 == 0 ? (vec[mid] + vec[mid-1]) / 2 : vec[mid];
}

Vector getMeanPos(vector<Vector> positions){
    int cnt = 0;
    Vector pos(4);
    for(int i=0; i < positions.size(); i++){
        pos[0] += positions[i][0];
        pos[1] += positions[i][1];
        pos[2] += positions[i][2];
        cnt++;
    }
    
    if(cnt==0){
        pos[3] = 0.0;
        return pos;
    }
    
    pos[0] /= cnt;
    pos[1] /= cnt;
    pos[2] /= cnt;
    
    pos[3] = 1.0;
    
    return pos;
}

class CtrlThread: public PeriodicThread
{
protected:
    BufferedPort<Bottle> inSkeletonPort;
    BufferedPort<ImageOf<PixelRgbFloat>>    cartesianImagePort;
    BufferedPort<ImageOf<PixelFloat>>   depthImagePort;
    BufferedPort<Bottle> targetPort;
    BufferedPort<Bottle>    ppsOutPort;

    Network yarp;
    
    bool                    connectedPPS;

    // camera parameters
    float focalLengthX = 918.4766845703125;
    float focalLengthY = 916.39385986328125;
    float principalPointX = 639.180908203125;
    float principalPointY = 342.84576416015625;
    
    float big_kpt_size = 0.02;
    float small_kpt_size = 0.002;
    
    /* iCub forward kinematics. */
    unique_ptr<ForwardKinematics> fk_;

    /* Calibratrion matrix. */
    Matrix calibration_matrix_;
    
    Port                    portToGui;
    yarp::os::Bottle        cmdGui;
    map<string, vector<Vector>> bodyParts;
    map<string, int> bodyCounter;
    map<string, int> bodyFailCounter;
    Stamp       ts;

public:
    CtrlThread() : PeriodicThread(1.0) { }

    virtual bool threadInit()
    {
        bool ret=true;
        ret=inSkeletonPort.open("/skeletonPosition/in");
        // yarp.connect("/yarpOpenPose/target:o", "/skeletonPosition/in");
        // ret = ret && cartesianImagePort.open("/skeletonPosition/cartesian:i");
        // yarp.connect("/SFM/world/cartesian:o", "/skeletonPosition/cartesian:i");

        yarp.connect("/keypointsDetector/keypoints", "/skeletonPosition/in");
        ret = ret && depthImagePort.open("/skeletonPosition/depthImage:i");

        yarp.connect("/depthCamera/depthImage:o", "/skeletonPosition/depthImage:i");

        ret = ret && targetPort.open("/skeletonPosition/out");
        
        // TODO: load params from some file
        string log_name_ = "skeletonPosition";
        string robot_name = "icub";
        string eye_version = "v2";
        string path = "/home/icub/data/eMc.txt";
        /* Try to open the file containing the calibration matrix. */
        std::ifstream in(path);
        if (!in.is_open())
        {
            yError() << "Error: cannot open the calibration matrix file " + path + ".";
            return false;
        }
        try
        {
            vpHomogeneousMatrix visp_H;
            visp_H.load(in);

            if (!visp_H.isAnHomogeneousMatrix())
            {
                yError() << log_name_ + "::configure(). The calibration matrix is not a valid homogeneous matrix";

                return false;
            }

            /* Store in a yarp::sig::Matrix for later use. */
            calibration_matrix_.resize(4, 4);
            for (std::size_t i = 0; i < 4; i++)
                for (std::size_t j = 0; j < 4; j++)
                    calibration_matrix_(i, j) = visp_H[i][j];

            yInfo() << log_name_ + "::configure(). Info: the loaded calibration matrix is :";
            yInfo() << log_name_ + "::configure(). Info: " + calibration_matrix_.toString();
        }
        catch (const vpException& e)
        {
            yError() << log_name_ + "::configure(). Error while reading the calibration matrix file " + path + ".";
            yError() << log_name_ + "::configure(). Error is: " + e.getStringMessage() + ".";

            return false;
        }

        fk_ = make_unique<ForwardKinematics>(robot_name, eye_version);
        string name = "skeleton";
        ppsOutPort.open(("/"+name+"/visuoTactileWrapper/objects:o").c_str());
        std::string visuoTactileWrapper_inport = "/visuoTactileWrapper/sensManager:i";
        connectedPPS = yarp::os::Network::connect(ppsOutPort.getName().c_str(), visuoTactileWrapper_inport);
        if (!connectedPPS)
            yError("[%s] Unable to connect to %s port", name.c_str(),visuoTactileWrapper_inport.c_str());
        else
            yInfo("[%s] Connected to %s port", name.c_str(), visuoTactileWrapper_inport.c_str());
        
        
        //**** visualizing human skeleton iCubGui ***************************
        string port2iCubGui = "/" + name + "/gui:o";
        if (!portToGui.open(port2iCubGui.c_str())) {
           yError("[%s] Unable to open port << port2iCubGui << endl", name.c_str());
        }
        std::string portGuiObject = "/iCubGui/objects";     // World frame
        yarp::os::Network::connect(port2iCubGui.c_str(), portGuiObject.c_str());

        cmdGui.clear();
        cmdGui.addString("reset");
        portToGui.write(cmdGui);
        
        ts.update();
        
        return ret;
    }
    /**************************************************/

    Vector get3DPositionInCameraFrame(int u, int v, float depth){
        Vector position(4);
        position[0] = (float)(depth * (float)u - principalPointX * depth) / focalLengthX;
        position[1] = (float)(depth * (float)v - principalPointY * depth) / focalLengthY;
        position[2] = depth;
        position[3] = 1; // add homogenous part
        return position;
    }

    Matrix getTransformationToCamera(){
        Matrix root_to_center = fk_->ee_pose();

        Matrix root_to_camera = root_to_center * calibration_matrix_;

        return root_to_camera;
    }

    virtual void run()
    {   
        // cartesian image is published less frequent than skeleton, so we first read 
        // cart. im. and only after that we check for last published skeleton
        // ImageOf<PixelRgbFloat> *cartesianImage = cartesianImagePort.read(false);

        ImageOf<PixelFloat> *depthImage = depthImagePort.read(false);

        if (depthImage != NULL){
            Bottle *skeleton = inSkeletonPort.read(false);
            if (skeleton != NULL){
                Bottle &target = ppsOutPort.prepare();
                target.clear();
                yInfo() << "Got data, lets process it..";

                // pass some datum ptr - desired data are in nested List
                Bottle *positions = skeleton;
                // procces actual data
                Matrix transformationRS = getTransformationToCamera();
                for(int person = 0; person < positions->size(); person++){
                    
                    Value person_val = positions->get(person);
                    Bottle *person_bot = person_val.asList();

                    Bottle &peopleList = target.addList();
                    for (int i = 0; i < person_bot->size(); i++){
                        Value val = person_bot->get(i);
                        Bottle *pos = val.asList();
                        string body_part = pos->get(0).asString();
                        double x_pos = pos->get(1).asDouble();
                        double y_pos = pos->get(2).asDouble();
                        double conf = pos->get(3).asDouble();
                        if ((conf < 0.3 && i < 26) || (conf < 0.07 && i > 25)){
                            bodyFailCounter[body_part] += 1;
                            continue;
                        }
                        // !!!!!!!!!!!!!!! DONT FORWARD FACE KEYPOINTS
                        // if (i > 26 && i < 95) continue;
                        // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        Bottle &partList = peopleList.addList();
                        // propagate body part
                        // get XYZ coords of current body part
                        int u = (int)(x_pos);
                        int v = (int)(y_pos);
                        if (u < 0 || u >= 1280 || v < 0 || v >= 720){
                            bodyFailCounter[body_part] += 1;
                            continue;
                        }
                        // PixelRgbFloat target_coord = cartesianImage->pixel(u, v);
                        float kpt_size;
                        if (i < 27){
                            kpt_size = big_kpt_size;
                        }
                        else{
                            kpt_size = small_kpt_size;
                        }
                        float depth = 0;
                        int cnt = 0;
                        for(int i_ = -1; i_ < 2; i_++){
                            for(int j_ = -1; j_ < 2; j_++){
                                int ii = u + i_;
                                int jj = v + j_;
                                if (ii < 0 || jj < 0 || ii >= 1280 || jj >= 720){
                                    continue;
                                }
                                depth += depthImage->pixel(ii, jj);
                                cnt++;
                            }
                        }
                        depth = depth / (float) cnt;
                        if (depth < 0.2){
                            bodyFailCounter[body_part] += 1;
                            continue;
                        }
                        // float depth = depthImage->pixel(u, v);
                        // Vector target_coord = get3DPositionInCameraFrame(u, v, depth);
                        int x_neighborhood = (int) floor(focalLengthX * 2 * kpt_size / depth);
                        int y_neighborhood = (int) floor(focalLengthY * 2 * kpt_size / depth);
                        int index_x = (int) floor(x_neighborhood / 2);
                        int index_y = (int) floor(y_neighborhood / 2);
                        Vector position(4, 0.0);

                        //for (int ui = 0; ui < x_neighborhood; ui++){
                        //    for (int vi = 0; vi < y_neighborhood; vi++){
                        //        int temp_u = u + ui - index_x;
                        //        int temp_v = v + vi - index_y;

                        //        if (temp_u < 0 || temp_u >= 1280 || temp_v < 0 || temp_v >= 720){
                        //            continue;
                        //        }
                        //        float temp_depth = depthImage->pixel(temp_u, temp_v);
                        //        Vector temp = get3DPositionInCameraFrame(temp_u, temp_v, temp_depth);
                        //        position[0] += temp[0];
                        //        position[1] += temp[1];
                        //        position[2] += temp[2];
                        //        position[3]++;
                        //    }
                        //}
                        //if (position[3] > 0){
                        //    position[0] /= position[3];
                        //    position[1] /= position[3];
                        //    position[2] /= position[3];
                        //    position[3] /= position[3];
                        //}
                        //else{
                        //    yWarning() << "Extracting position failed!";
                        //}
                        
                        Vector median_position(4, 0.0);
                        
                        vector<float> x_position;
                        vector<float> y_position;
                        vector<float> z_position;
                        int med_cnt = 0;
                        for (int ui = 0; ui < x_neighborhood; ui++){
                            for (int vi = 0; vi < y_neighborhood; vi++){
                                int temp_u = u + ui - index_x;
                                int temp_v = v + vi - index_y;

                                if (temp_u < 0 || temp_u >= 1280 || temp_v < 0 || temp_v >= 720){
                                    continue;
                                }
                                float temp_depth = depthImage->pixel(temp_u, temp_v);
                                if (temp_depth <= 0.2){
                                    continue;
                                }
                                Vector temp = get3DPositionInCameraFrame(temp_u, temp_v, temp_depth);
                                x_position.push_back(temp[0]);
                                y_position.push_back(temp[1]);
                                z_position.push_back(temp[2]);
                                med_cnt++;
                            }
                        }
                        
                        if (med_cnt == 0){
                            bodyFailCounter[body_part] += 1;
                            continue;
                        }
                        
                        bodyFailCounter[body_part] = 0;
                        
                        median_position[0] = median(x_position);
                        median_position[1] = median(y_position);
                        median_position[2] = median(z_position);
                        median_position[3] = 1.0;
                        
                        // position = transformationRS * position;
                        position = transformationRS * median_position;
                        partList.addDouble(position[0]);
                        partList.addDouble(position[1]);
                        partList.addDouble(position[2]);
                        if (body_part == "RWrist"){
                            yInfo() << "Position of " << body_part << " is " << "(" << position[0] << ", " << position[1] << ", " << position[2] << ") with confidence = " << conf;
                        }
                        // propagate the position, valence and confidence
                        float valence = (i < 27) ? (0.07 / 2.0) : (0.02 /2.0); 
                        partList.addDouble(valence);
                        partList.addDouble(1.0 + 1.0*(1-conf) - 1.0);
                        map<string, int>::iterator it;
                        it = bodyCounter.find(body_part);
                        if (it != bodyCounter.end()){
                            bodyCounter[body_part] = (bodyCounter[body_part] + 1) % 5; 
                        }
                        else{
                            bodyCounter[body_part] = 0;
                        }
                        map<string, vector<Vector>>::iterator it_;
                        it_ = bodyParts.find(body_part);
                        if (it_ != bodyParts.end()){
                            if (bodyParts[body_part].size() < 5){
                                bodyParts[body_part].push_back(position);
                            }
                            else{
                                bodyParts[body_part][bodyCounter[body_part]] = position;
                            }
                        }
                        else{
                            vector<Vector> part_vec{position};
                            bodyParts[body_part] = part_vec;
                        }
                    }
                }
                ts.update();
                // publish output with human parts and its XYZ coordinates + confidence (propagated from input skeleton)
                ppsOutPort.setEnvelope(ts);
                ppsOutPort.write();
                drawBodyGui(bodyParts, bodyFailCounter);
            }
        }
    }

    virtual void threadRelease()
    {
        targetPort.close();
        inSkeletonPort.close();
        cartesianImagePort.close();
        portToGui.close();
    }
    
    void    initShowBodySegGui(const string &segmentName, const string &color)
{
    cmdGui.clear();

    cmdGui.addString("trajectory");
    cmdGui.addString(segmentName.c_str());    // trajectory identifier
    cmdGui.addString("");               // trajectory name
    cmdGui.addInt(512);                 // max samples in circular queue
    cmdGui.addDouble(1200.0);             // lifetime of samples
    if (color =="red")             // color
    {
        cmdGui.addInt(255);cmdGui.addInt(0);cmdGui.addInt(0); //red
    }
    else if (color =="green")
    {
        cmdGui.addInt(0);cmdGui.addInt(255);cmdGui.addInt(0); //green
    }
    else if (color =="blue")
    {
        cmdGui.addInt(0);cmdGui.addInt(0);cmdGui.addInt(255); //blue
    }
    else if (color =="purple")
    {
        cmdGui.addInt(255);cmdGui.addInt(0);cmdGui.addInt(255); //purple
    }
    else if (color =="yellow")
    {
        cmdGui.addInt(255);cmdGui.addInt(255);cmdGui.addInt(0); //yellow
    }
    cmdGui.addDouble(1.0);             // alpha [0,1]
    cmdGui.addDouble(5.0);             // line width

    portToGui.write(cmdGui);
}

void    updateBodySegGui(const vector<Vector> &segment, const string &segmentName)
{
    if (segment.size()>0)
    {
        for (int i=0; i<segment.size(); i++)
        {
            cmdGui.clear();

            cmdGui.addString("addpoint");
            cmdGui.addString(segmentName.c_str());                    // trajectory identifier
            cmdGui.addDouble(1000.0*segment[i][0]);      // posX [mm]
            cmdGui.addDouble(1000.0*segment[i][1]);      // posY [mm]
            cmdGui.addDouble(1000.0*segment[i][2]);      // posZ [mm]

            portToGui.write(cmdGui);
        }
    }
}

void    deleteBodySegGui(const string &segmentName)
{
    cmdGui.clear();
    cmdGui.addString("delete");
    cmdGui.addString(segmentName.c_str());
    portToGui.write(cmdGui);
}

bool    drawBodyGui(map<string, vector<Vector>> bodyParts, map<string, int> bodyFailCounter)
{
    vector<string> upperTrajectory{"RWrist", "RElbow", "RShoulder", "Neck", "LShoulder", "LElbow", "LWrist"};
    vector<string> spineTrajectory{"Neck", "Hip", "LHip", "Hip", "RHip"};
    vector<string> leftHandTrajectory{"LWrist", "LThumpMcp", "LWrist", "LIndexFingerPip", "LWrist", "LMiddleFingerPip", "LWrist", "LRingFingerPip", "LWrist", "LPinkyPip"};
    vector<string> rightHandTrajectory{"RWrist", "RThumpMcp", "RWrist", "RIndexFingerPip", "RWrist", "RMiddleFingerPip", "RWrist", "RRingFingerPip", "RWrist", "RPinkyPip"};
    deleteBodySegGui("upper");
    deleteBodySegGui("spine");
    deleteBodySegGui("left_hand");
    deleteBodySegGui("right_hand");
    map<string, vector<Vector>>::iterator it;
    initShowBodySegGui("upper","blue");
    initShowBodySegGui("spine","red");
    initShowBodySegGui("left_hand", "blue");
    initShowBodySegGui("right_hand", "blue");
    vector<Vector> segmentUpper;
    for(int i=0; i < 7; i++){
        it = bodyParts.find(upperTrajectory[i]);
        if (it != bodyParts.end()){
            Vector pos = getMeanPos(bodyParts[upperTrajectory[i]]);
            if (pos[3] != 0.0 && bodyFailCounter[upperTrajectory[i]] < 5)
                segmentUpper.push_back(pos);
        }
    }
    updateBodySegGui(segmentUpper, "upper");
    vector<Vector> segmentSpine;
    for(int i=0; i < 5; i++){
        it = bodyParts.find(spineTrajectory[i]);
        if (it != bodyParts.end()){
            Vector pos = getMeanPos(bodyParts[spineTrajectory[i]]);
            if (pos[3] != 0.0 && bodyFailCounter[spineTrajectory[i]] < 5)
                segmentSpine.push_back(pos);
        }
    }
    updateBodySegGui(segmentSpine, "spine");
    
    vector<Vector> segmentLeftHand;
    for(int i=0; i < 10; i++){
        it = bodyParts.find(leftHandTrajectory[i]);
        if (it != bodyParts.end()){
            Vector pos = getMeanPos(bodyParts[leftHandTrajectory[i]]);
            if (pos[3] != 0.0 && bodyFailCounter[leftHandTrajectory[i]] < 5)
                segmentLeftHand.push_back(pos);
        }
    }
    updateBodySegGui(segmentLeftHand, "left_hand");
    
    vector<Vector> segmentRightHand;
    for(int i=0; i < 10; i++){
        it = bodyParts.find(rightHandTrajectory[i]);
        if (it != bodyParts.end()){
            Vector pos = getMeanPos(bodyParts[rightHandTrajectory[i]]);
            if (pos[3] != 0.0 && bodyFailCounter[rightHandTrajectory[i]] < 5)
                segmentRightHand.push_back(pos);
        }
    }
    updateBodySegGui(segmentRightHand, "right_hand");
    
    return true;
}

    
};


class CtrlModule: public RFModule
{
protected:
    CtrlThread thr;

public:
    virtual bool configure(ResourceFinder &rf)
    {
        // retrieve command line options
        double period=rf.check("period",Value(0.02)).asDouble();

        // set the thread period in [s]
        thr.setPeriod(period);

        return thr.start();
    }

    virtual bool close()
    {
        thr.stop();
        return true;
    }

    virtual double getPeriod()
    {
        return 1.0;
    }

    virtual bool updateModule()
    {
        return true;
    }
};



int main(int argc, char *argv[])
{
    Network yarp;
    if (!yarp.checkNetwork())
    {
        yError()<<"YARP doesn't seem to be available";
        return EXIT_FAILURE;
    }

    ResourceFinder rf;
    rf.configure(argc,argv);

    CtrlModule mod;
    return mod.runModule(rf);
}

