#!/usr/bin/python3
import sys

import numpy as np
import cv2
import time
import math
import platform
import os
import torch
from argparse import ArgumentParser
import yarp
# path to directory with AlphaPose detector
sys.path.append("/home/icub/Documents/detectors/AlphaPose/")
from alphapose.utils.transforms import get_func_heatmap_to_coord
from alphapose.utils.pPose_nms import pose_nms
from alphapose.utils.presets import SimpleTransform
from alphapose.utils.transforms import flip, flip_heatmap
from alphapose.models import builder
from alphapose.utils.config import update_config
from detector.apis import get_detector
from alphapose.utils.vis import getTime

import mediapipe as mp

WIDTH = 1280
HEIGHT = 720


class DetectionLoader():
    def __init__(self, detector, cfg, opt):
        self.cfg = cfg
        self.opt = opt
        self.device = opt.device
        self.detector = detector

        self._input_size = cfg.DATA_PRESET.IMAGE_SIZE
        self._output_size = cfg.DATA_PRESET.HEATMAP_SIZE

        self._sigma = cfg.DATA_PRESET.SIGMA

        pose_dataset = builder.retrieve_dataset(self.cfg.DATASET.TRAIN)
        if cfg.DATA_PRESET.TYPE == 'simple':
            self.transformation = SimpleTransform(
                pose_dataset, scale_factor=0,
                input_size=self._input_size,
                output_size=self._output_size,
                rot=0, sigma=self._sigma,
                train=False, add_dpg=False, gpu_device=self.device)

        self.image = (None, None, None, None)
        self.det = (None, None, None, None, None, None, None)
        self.pose = (None, None, None, None, None, None, None)

    def process(self, im_name, image):
        # start to pre process images for object detection
        self.image_preprocess(im_name, image)
        # start to detect human in images
        self.image_detection()
        # start to post process cropped human image for pose estimation
        self.image_postprocess()
        return self

    def image_preprocess(self, im_name, image):
        # expected image shape like (1,3,h,w) or (3,h,w)
        img = self.detector.image_preprocess(image)
        if isinstance(img, np.ndarray):
            img = torch.from_numpy(img)
        # add one dimension at the front for batch if image shape (3,h,w)
        if img.dim() == 3:
            img = img.unsqueeze(0)
        orig_img = image  # scipy.misc.imread(im_name_k, mode='RGB') is depreciated
        im_dim = orig_img.shape[1], orig_img.shape[0]

        im_name = os.path.basename(im_name)

        with torch.no_grad():
            im_dim = torch.FloatTensor(im_dim).repeat(1, 2)
        self.image = (img, orig_img, im_name, im_dim)

    def image_detection(self):
        imgs, orig_imgs, im_names, im_dim_list = self.image
        if imgs is None:
            self.det = (None, None, None, None, None, None, None)
            return

        with torch.no_grad():
            dets = self.detector.images_detection(imgs, im_dim_list)
            if isinstance(dets, int) or dets.shape[0] == 0:
                self.det = (orig_imgs, im_names, None, None, None, None, None)
                return
            if isinstance(dets, np.ndarray):
                dets = torch.from_numpy(dets)
            dets = dets.cpu()
            boxes = dets[:, 1:5]
            scores = dets[:, 5:6]
            ids = torch.zeros(scores.shape)
        boxes = boxes[dets[:, 0] == 0]
        if isinstance(boxes, int) or boxes.shape[0] == 0:
            self.det = (orig_imgs, im_names, None, None, None, None, None)
            return
        inps = torch.zeros(boxes.size(0), 3, *self._input_size)
        cropped_boxes = torch.zeros(boxes.size(0), 4)
        self.det = (orig_imgs, im_names, boxes, scores[dets[:, 0] == 0], ids[dets[:, 0] == 0], inps, cropped_boxes)

    def image_postprocess(self):
        with torch.no_grad():
            (orig_img, im_name, boxes, scores, ids, inps, cropped_boxes) = self.det
            if orig_img is None:
                self.pose = (None, None, None, None, None, None, None)
                return
            if boxes is None or boxes.nelement() == 0:
                self.pose = (None, orig_img, im_name, boxes, scores, ids, None)
                return

            for i, box in enumerate(boxes):
                inps[i], cropped_box = self.transformation.test_transform(orig_img, box)
                cropped_boxes[i] = torch.FloatTensor(cropped_box)

            self.pose = (inps, orig_img, im_name, boxes, scores, ids, cropped_boxes)

    def read(self):
        return self.pose


class DataWriter():
    def __init__(self, cfg, opt):
        self.cfg = cfg
        self.opt = opt

        self.eval_joints = list(range(cfg.DATA_PRESET.NUM_JOINTS))
        self.heatmap_to_coord = get_func_heatmap_to_coord(cfg)
        self.item = (None, None, None, None, None, None, None)

        loss_type = self.cfg.DATA_PRESET.get('LOSS_TYPE', 'MSELoss')
        num_joints = self.cfg.DATA_PRESET.NUM_JOINTS
        if loss_type == 'MSELoss':
            self.vis_thres = [0.4] * num_joints
        elif 'JointRegression' in loss_type:
            self.vis_thres = [0.05] * num_joints
        elif loss_type == 'Combined':
            if num_joints == 68:
                hand_face_num = 42
            else:
                hand_face_num = 110
            self.vis_thres = [0.4] * (num_joints - hand_face_num) + [0.05] * hand_face_num

        self.use_heatmap_loss = (self.cfg.DATA_PRESET.get('LOSS_TYPE', 'MSELoss') == 'MSELoss')

    def start(self):
        # start to read pose estimation results
        return self.update()

    def update(self):
        norm_type = self.cfg.LOSS.get('NORM_TYPE', None)
        hm_size = self.cfg.DATA_PRESET.HEATMAP_SIZE

        # get item
        (boxes, scores, ids, hm_data, cropped_boxes, orig_img, im_name) = self.item
        if orig_img is None:
            return None
        # image channel RGB->BGR
        orig_img = np.array(orig_img, dtype=np.uint8)[:, :, ::-1]
        self.orig_img = orig_img
        if boxes is None or len(boxes) == 0:
            return None
        else:
            # location prediction (n, kp, 2) | score prediction (n, kp, 1)
            assert hm_data.dim() == 4
            if hm_data.size()[1] == 136:
                self.eval_joints = [*range(0, 136)]
            elif hm_data.size()[1] == 26:
                self.eval_joints = [*range(0, 26)]
            elif hm_data.size()[1] == 133:
                self.eval_joints = [*range(0, 133)]
            pose_coords = []
            pose_scores = []

            for i in range(hm_data.shape[0]):
                bbox = cropped_boxes[i].tolist()
                if isinstance(self.heatmap_to_coord, list):
                    pose_coords_body_foot, pose_scores_body_foot = self.heatmap_to_coord[0](
                        hm_data[i][self.eval_joints[:-110]], bbox, hm_shape=hm_size, norm_type=norm_type)
                    pose_coords_face_hand, pose_scores_face_hand = self.heatmap_to_coord[1](
                        hm_data[i][self.eval_joints[-110:]], bbox, hm_shape=hm_size, norm_type=norm_type)
                    pose_coord = np.concatenate((pose_coords_body_foot, pose_coords_face_hand), axis=0)
                    pose_score = np.concatenate((pose_scores_body_foot, pose_scores_face_hand), axis=0)
                else:
                    pose_coord, pose_score = self.heatmap_to_coord(hm_data[i][self.eval_joints], bbox, hm_shape=hm_size,
                                                                   norm_type=norm_type)
                pose_coords.append(torch.from_numpy(pose_coord).unsqueeze(0))
                pose_scores.append(torch.from_numpy(pose_score).unsqueeze(0))
            preds_img = torch.cat(pose_coords)
            preds_scores = torch.cat(pose_scores)

            boxes, scores, ids, preds_img, preds_scores, pick_ids = \
                pose_nms(boxes, scores, ids, preds_img, preds_scores, self.opt.min_box_area,
                         use_heatmap_loss=self.use_heatmap_loss)

            _result = []
            for k in range(len(scores)):
                _result.append(
                    {
                        'keypoints': preds_img[k],
                        'kp_score': preds_scores[k],
                        'proposal_score': torch.mean(preds_scores[k]) + scores[k] + 1.25 * max(preds_scores[k]),
                        'idx': ids[k],
                        'bbox': [boxes[k][0], boxes[k][1], boxes[k][2] - boxes[k][0], boxes[k][3] - boxes[k][1]]
                    }
                )

            result = {
                'imgname': im_name,
                'result': _result
            }

            if hm_data.size()[1] == 49:
                from alphapose.utils.vis import vis_frame_dense as vis_frame
            elif self.opt.vis_fast:
                from alphapose.utils.vis import vis_frame_fast as vis_frame
            else:
                from alphapose.utils.vis import vis_frame
            self.vis_frame = vis_frame

        return result

    def save(self, boxes, scores, ids, hm_data, cropped_boxes, orig_img, im_name):
        self.item = (boxes, scores, ids, hm_data, cropped_boxes, orig_img, im_name)


class SingleImageAlphaPose():
    def __init__(self, args, cfg):
        self.args = args
        self.cfg = cfg

        # Load pose model
        self.pose_model = builder.build_sppe(cfg.MODEL, preset_cfg=cfg.DATA_PRESET)

        print(f'Loading pose model from {args.checkpoint}...')
        self.pose_model.load_state_dict(torch.load(args.checkpoint, map_location=args.device))
        self.pose_dataset = builder.retrieve_dataset(cfg.DATASET.TRAIN)

        self.pose_model.to(args.device)
        self.pose_model.eval()

        self.det_loader = DetectionLoader(get_detector(self.args), self.cfg, self.args)

    def process(self, im_name, image):
        # Init data writer
        self.writer = DataWriter(self.cfg, self.args)

        runtime_profile = {
            'dt': [],
            'pt': [],
            'pn': []
        }
        pose = None
        try:
            start_time = getTime()
            with torch.no_grad():
                (inps, orig_img, im_name, boxes, scores, ids, cropped_boxes) = self.det_loader.process(im_name,
                                                                                                       image).read()
                if orig_img is None:
                    raise Exception("no image is given")
                if boxes is None or boxes.nelement() == 0:
                    if self.args.profile:
                        ckpt_time, det_time = getTime(start_time)
                        runtime_profile['dt'].append(det_time)
                    self.writer.save(None, None, None, None, None, orig_img, im_name)
                    if self.args.profile:
                        ckpt_time, pose_time = getTime(ckpt_time)
                        runtime_profile['pt'].append(pose_time)
                    pose = self.writer.start()
                    if self.args.profile:
                        ckpt_time, post_time = getTime(ckpt_time)
                        runtime_profile['pn'].append(post_time)
                else:
                    if self.args.profile:
                        ckpt_time, det_time = getTime(start_time)
                        runtime_profile['dt'].append(det_time)
                    # Pose Estimation
                    inps = inps.to(self.args.device)
                    if self.args.flip:
                        inps = torch.cat((inps, flip(inps)))
                    hm = self.pose_model(inps)
                    if self.args.flip:
                        hm_flip = flip_heatmap(hm[int(len(hm) / 2):], self.pose_dataset.joint_pairs, shift=True)
                        hm = (hm[0:int(len(hm) / 2)] + hm_flip) / 2
                    if self.args.profile:
                        ckpt_time, pose_time = getTime(ckpt_time)
                        runtime_profile['pt'].append(pose_time)
                    hm = hm.cpu()
                    self.writer.save(boxes, scores, ids, hm, cropped_boxes, orig_img, im_name)
                    pose = self.writer.start()
                    if self.args.profile:
                        ckpt_time, post_time = getTime(ckpt_time)
                        runtime_profile['pn'].append(post_time)

            if self.args.profile:
                print(
                    'det time: {dt:.4f} | pose time: {pt:.4f} | post processing: {pn:.4f}'.format(
                        dt=np.mean(runtime_profile['dt']), pt=np.mean(runtime_profile['pt']),
                        pn=np.mean(runtime_profile['pn']))
                )
            # print('===========================> Finish Model Running.')
        except Exception as e:
            print(repr(e))
            print('An error as above occurs when processing the images, please check it')
            pass
        except KeyboardInterrupt:
            print('===========================> Finish Model Running.')

        return pose

    def getImg(self):
        return self.writer.orig_img

    def vis(self, image, pose):
        if pose is not None:
            image = self.writer.vis_frame(image, pose, self.writer.opt, self.writer.vis_thres)
        return image

    def writeJson(self, final_result, outputpath, form='coco', for_eval=False):
        from alphapose.utils.pPose_nms import write_json
        write_json(final_result, outputpath, form=form, for_eval=for_eval)
        print("Results have been written to json.")


def main():
    parser = ArgumentParser(description='AlphaPose Single-Image Demo')
    parser.add_argument('--cfg',
                        default='/home/icub/Documents/detectors/AlphaPose/configs/halpe_136/resnet/256x192_res50_lr1e-3_2x-dcn-combined.yaml',
                        type=str,
                        help='experiment configure file name')
    parser.add_argument('--checkpoint',
                        default='/home/icub/Documents/detectors/AlphaPose/pretrained_models/multi_domain_fast50_dcn_combined_256x192.pth',
                        type=str,
                        help='checkpoint file name')
    parser.add_argument('--detector', dest='detector',
                        help='detector name', default="yolo")
    parser.add_argument('--vis', default=True, action='store_true',
                        help='visualize image')
    parser.add_argument('--showbox', default=False, action='store_true',
                        help='visualize human bbox')
    parser.add_argument('--profile', default=False, action='store_true',
                        help='add speed profiling at screen output')
    parser.add_argument('--format', type=str,
                        help='save in the format of cmu or coco or openpose, option: coco/cmu/open')
    parser.add_argument('--min_box_area', type=int, default=0,
                        help='min box area to filter out')
    parser.add_argument('--eval', dest='eval', default=False, action='store_true',
                        help='save the result json as coco format, using image index(int) instead of image name(str)')
    parser.add_argument('--gpus', type=str, dest='gpus', default="0",
                        help='choose which cuda device to use by index and input comma to use multi gpus, e.g. 0,1,2,3. (input -1 for cpu only)')
    parser.add_argument('--flip', default=False, action='store_true',
                        help='enable flip testing')
    parser.add_argument('--debug', default=False, action='store_true',
                        help='print detail information')
    parser.add_argument('--vis_fast', dest='vis_fast',
                        help='use fast rendering', action='store_true', default=True)
    """----------------------------- Tracking options -----------------------------"""
    parser.add_argument('--pose_flow', dest='pose_flow',
                        help='track humans in video with PoseFlow', action='store_true', default=False)
    parser.add_argument('--pose_track', dest='pose_track',
                        help='track humans in video with reid', action='store_true', default=False)

    args = parser.parse_args()
    cfg = update_config(args.cfg)

    args.gpus = [int(args.gpus[0])] if torch.cuda.device_count() >= 1 else [-1]
    args.device = torch.device("cuda:" + str(args.gpus[0]) if args.gpus[0] >= 0 else "cpu")
    args.tracking = args.pose_track or args.pose_flow or args.detector == 'tracker'
    # Initialise YARP
    yarp.Network.init()

    # Create a port and connect it to the realsense
    input_port = yarp.Port()
    input_port.open("/keypointsDetector/rgbImage:i")
    # yarp.Network.connect("/yarpOpenPose/image:o", input_port.getName())
    yarp.Network.connect("/depthCamera/rgbImage:o", input_port.getName())

    # Create numpy array to receive the image and the YARP image wrapped around it
    img_array = np.zeros((HEIGHT, WIDTH, 3), dtype=np.uint8)
    yarp_image = yarp.ImageRgb()
    yarp_image.resize(WIDTH, HEIGHT)
    yarp_image.setExternal(img_array, img_array.shape[1], img_array.shape[0])

    # Create output visualize port and connect it to the yarpview figure
    output_port = yarp.Port()
    output_port.open("/keypointsDetector/rgbImage:o")
    yarp.Network.connect(output_port.getName(), "/detectedKeypoints")

    # Create output keypoints port and connect it further
    keypoints_port = yarp.Port()
    keypoints_port.open("/keypointsDetector/keypoints")
    yarp.Network.connect(keypoints_port.getName(),
                         "/keypoints_out")  # TODO: specify for 2D->3D position script later on

    detector = SingleImageAlphaPose(args, cfg)

    # Prepare mediapipe hand detector
    mp_hands = mp.solutions.hands
    mp_drawing = mp.solutions.drawing_utils
    mp_drawing_styles = mp.solutions.drawing_styles

    hands = mp_hands.Hands(static_image_mode=False,
                           max_num_hands=2,
                           min_detection_confidence=0.2,
                           min_tracking_confidence=0.4)

    halpeParts = ["Nose", "LEye", "REye", "LEar", "REar", "LShoulder", "RShoulder", "LElbow", "RElbow", "LWrist",
                  "RWrist", "LHip", "RHip", "LKnee", "RKnee", "LAnkle", "RAnkle", "Head", "Neck", "Hip", "LBigToe",
                  "RBigToe", "LSmallToe", "RSmallToe", "LHeel", "RHeel", "RJaw_0", "RJaw_1", "RJaw_2", "RJaw_3",
                  "RJaw_4", "RJaw_5", "Chin_0", "Chin_1", "Chin_2", "Chin_3", "Chin_4", "LJaw_5", "LJaw_4", "LJaw_3",
                  "LJaw_2", "LJaw_1", "LJaw_0", "Forehead_0", "Forehead_1", "Forehead_2", "Forehead_3", "Forehead_4",
                  "Forehead_5", "Forehead_6", "Forehead_7", "Forehead_8", "Forehead_9", "FNose_0", "FNose_1", "FNose_2",
                  "FNose_3", "FNose_4", "FNose_5", "FNose_6", "FNose_7", "FNose_8", "FREye_0", "FREye_1", "FREye_2",
                  "FREye_3", "FREye_4", "FREye_5", "FLEye_0", "FLEye_1", "FLEye_2", "FLEye_3", "FLEye_4", "FLEye_5",
                  "Lips_0", "Lips_1", "Lips_2", "Lips_3", "Lips_4", "Lips_5", "Lips_6", "Lips_7", "Lips_8", "Lips_9",
                  "Lips_10", "Lips_11", "Mouth_0", "Mouth_1", "Mouth_2", "Mouth_3", "Mouth_4", "Mouth_5", "Mouth_6",
                  "Mouth_7", "LHWrist", "LThumbCmc", "LThumbMcp", "LThumbIp", "LThumbTip", "LIndexFingerMcp",
                  "LIndexFingerPip", "LIndexFingerDip", "LIndexFingerTip", "LMiddleFingerMcp", "LMiddleFingerPip",
                  "LMiddleFingerDip", "LMiddleFingerTip", "LRingFingerMcp", "LRingFingerPip", "LRingFingerDip",
                  "LRingFingerTip", "LPinkyMcp", "LPinkyDip", "LPinkyPip", "LPinkyTip", "RHWrist", "RThumbCmc",
                  "RThumbMcp", "RThumbIp", "RThumbTip", "RIndexFingerMcp", "RIndexFingerPip", "RIndexFingerDip",
                  "RIndexFingerTip", "RMiddleFingerMcp", "RMiddleFingerPip", "RMiddleFingerDip", "RMiddleFingerTip",
                  "RRingFingerMcp", "RRingFingerPip", "RRingFingerDip", "RRingFingerTip", "RPinkyMcp", "RPinkyDip",
                  "RPinkyPip", "RPinkyTip"]
    
    # Functions to process input image
    def hand_detection(im):
        # res = hands.process(cv2.flip(image, 1))
        # if res is not None:
        #     for h_landmarks in res.multi_hand_landmarks:
        #         for landmark in h_landmarks.landmark:
        #             landmark.x = 1 - landmark.x
        # TODO: OR - rather do not flip image and x coordinates but swap Left/Right label
        # ----> act like Left is Right and vice versa
        res = hands.process(im)
        right_hand_kpts = None
        left_hand_kpts = None
        # go through hand landmarks detected by mediapipe hand detector
        if res.multi_hand_landmarks:
            for idx, hand_landmarks in enumerate(res.multi_hand_landmarks):
                handedness = res.multi_handedness[idx]
                kpts_x = np.zeros(21)
                kpts_y = np.zeros(21)
                kpts_conf = np.zeros(21)

                # compute keypoints coordinates
                for i, landmark in enumerate(hand_landmarks.landmark):
                    x_coord = landmark.x * WIDTH
                    y_coord = landmark.y * HEIGHT

                    if 0 <= x_coord < WIDTH and 0 <= y_coord < HEIGHT:
                        kpts_x[i] = x_coord
                        kpts_y[i] = y_coord
                        kpts_conf[i] = 0.5
                if handedness.classification[0].label == "Right":
                    left_hand_kpts = np.zeros(3 * 21)
                    left_hand_kpts[0::3] = kpts_x
                    left_hand_kpts[1::3] = kpts_y
                    left_hand_kpts[2::3] = kpts_conf
                else:
                    right_hand_kpts = np.zeros(3 * 21)
                    right_hand_kpts[0::3] = kpts_x
                    right_hand_kpts[1::3] = kpts_y
                    right_hand_kpts[2::3] = kpts_conf

        return res, left_hand_kpts, right_hand_kpts

    def wholebody_detection(im):
        im_name = "current_image"       # just some image flag
        res = detector.process(im_name, im)
        return res
    
    print("Detector and ports prepared.")
    while True:
        # Read the data from the port into the image
        input_port.read(yarp_image)
        
        # convert numpy array image to cv2 image
        image = cv2.cvtColor(img_array, cv2.COLOR_BGR2RGB)

        # process image using both detectors
        pose = wholebody_detection(image)
        hand_results, left_hand_results, right_hand_results = hand_detection(image)

        if left_hand_results is not None:
            distance = 1000000
            closest_i = None
            if pose is not None:
                for i in range(len(pose['result'])):
                    curr_kpts = pose['result'][i]['keypoints'].numpy().T
                    x_kpts = curr_kpts[0]
                    y_kpts = curr_kpts[1]
                    l_wrist_x = x_kpts[9]
                    l_wrist_y = y_kpts[9]
                    l_wrist_x2 = left_hand_results[0]
                    l_wrist_y2 = left_hand_results[1]

                    dist = np.sqrt((l_wrist_x - l_wrist_x2)**2 + (l_wrist_y - l_wrist_y2)**2)
                    if dist < distance:
                        distance = dist
                        closest_i = i
                if closest_i is not None:
                    for i in range(int(len(left_hand_results) / 3)):
                        pose['result'][closest_i]['keypoints'][94 + i][0] = left_hand_results[3 * i]
                        pose['result'][closest_i]['keypoints'][94 + i][1] = left_hand_results[3 * i + 1]
        if right_hand_results is not None:
            distance = 1000000
            closest_i = None
            if pose is not None:
                for i in range(len(pose['result'])):
                    curr_kpts = pose['result'][i]['keypoints'].numpy().T
                    x_kpts = curr_kpts[0]
                    y_kpts = curr_kpts[1]
                    r_wrist_x = x_kpts[10]
                    r_wrist_y = y_kpts[10]
                    r_wrist_x2 = right_hand_results[0]
                    r_wrist_y2 = right_hand_results[1]

                    dist = np.sqrt((r_wrist_x - r_wrist_x2) ** 2 + (r_wrist_y - r_wrist_y2) ** 2)
                    if dist < distance:
                        distance = dist
                        closest_i = i
                if closest_i is not None:
                    for i in range(int(len(right_hand_results) / 3)):
                        pose['result'][closest_i]['keypoints'][115 + i][0] = right_hand_results[3 * i]
                        pose['result'][closest_i]['keypoints'][115 + i][1] = right_hand_results[3 * i + 1]

        # process the results (send keypoints and visualized image to ports)
        if args.vis:
            img_out = image
            img_out = detector.vis(img_out, pose)
            if hand_results.multi_hand_landmarks:
                for hand_landmarks in hand_results.multi_hand_landmarks:
                    mp_drawing.draw_landmarks(img_out, hand_landmarks, mp_hands.HAND_CONNECTIONS,
                                          mp_drawing_styles.get_default_hand_landmarks_style(),
                                          mp_drawing_styles.get_default_hand_connections_style())
            img_array[:] = cv2.cvtColor(img_out, cv2.COLOR_RGB2BGR)
            output_port.write(yarp_image)

        if pose is None:
            continue

        # find detection with highest score -> could be also "pick detections with score higher than THRESHOLD"
        best_score = 0
        best_i = None
        for i in range(len(pose['result'])):
            score = float(pose['result'][i]['proposal_score'])
            if score > best_score:
                best_score = score
                best_i = i

        best_kpts = pose['result'][best_i]['keypoints'].numpy().T
        x_kpts = best_kpts[0]
        y_kpts = best_kpts[1]
        score_kpts = pose['result'][best_i]['kp_score'].numpy()
        NUM_KPTS = 136
        bout = yarp.Bottle()
        kpts = bout.addList()
        for i in range(NUM_KPTS):
            kpt = kpts.addList()
            kpt.addString(halpeParts[i])
            kpt.addDouble(float(x_kpts[i]))
            kpt.addDouble(float(y_kpts[i]))
            kpt.addDouble(float(score_kpts[i]))
        keypoints_port.write(bout)

        # process keypoints of the highest score detection
        # for i in range(len(pose['result'])):
          #  best_kpts = pose['result'][i]['keypoints'].numpy().T
           # x_kpts = best_kpts[0]
            #y_kpts = best_kpts[1]
            #score_kpts = pose['result'][i]['kp_score'].numpy()
            #NUM_KPTS = 136
            #bout = yarp.Bottle()
            #kpts = bout.addList()
            #for i in range(NUM_KPTS):
            #    kpt = kpts.addList()
            #    kpt.addString(halpeParts[i])
            #    kpt.addDouble(float(x_kpts[i]))
            #    kpt.addDouble(float(y_kpts[i]))
            #    kpt.addDouble(float(score_kpts[i]))
            #keypoints_port.write(bout)

    # Cleanup
    input_port.close()
    output_port.close()
    keypoints_port.close()


#######
# command similar to
# python scripts/demo_inference.py --cfg configs/halpe_136/resnet/256x192_res50_lr1e-3_2x-dcn-combined.yaml --checkpoint pretrained_models/multi_domain_fast50_dcn_combined_256x192.pth
if __name__ == '__main__':
    main()
