#!/usr/bin/python3
import sys
# path to MMPose repository
sys.path.append("/home/icub/Documents/detectors/mmpose/")
import numpy
import skimage.feature
import yarp
import matplotlib.pyplot as plt
import cv2
import time
from argparse import ArgumentParser
# import mediapipe as mp
import warnings
from mmpose.apis import (inference_top_down_pose_model, init_pose_model,
                         process_mmdet_results, vis_pose_result)
from mmpose.datasets import DatasetInfo

try:
    from mmdet.apis import inference_detector, init_detector
    has_mmdet = True
except (ImportError, ModuleNotFoundError):
    has_mmdet = False

def main():
    # Prepare ArgumentPaserser
    parser = ArgumentParser()
    parser.add_argument('--det_config', default='/home/icub/Documents/detectors/mmpose/demo/mmdetection_cfg/faster_rcnn_r50_fpn_coco.py', help='Config file for detection')
    parser.add_argument('--det_checkpoint', default='https://download.openmmlab.com/mmdetection/v2.0/faster_rcnn/faster_rcnn_r50_fpn_1x_coco/faster_rcnn_r50_fpn_1x_coco_20200130-047c8118.pth', help='Checkpoint file for detection')
    parser.add_argument('--pose_config', default='/home/icub/Documents/detectors/mmpose/configs/wholebody/2d_kpt_sview_rgb_img/topdown_heatmap/coco-wholebody/hrnet_w48_coco_wholebody_384x288_dark_plus.py', help='Config file for pose')
    parser.add_argument('--pose_checkpoint', default='https://download.openmmlab.com/mmpose/top_down/hrnet/hrnet_w48_coco_wholebody_384x288_dark-f5726563_20200918.pth', help='Checkpoint file for pose')
    parser.add_argument(
        '--show',
        action='store_true',
        default=True,
        help='whether to show img')
    parser.add_argument(
        '--device', default='cuda:0', help='Device used for inference')
    parser.add_argument(
        '--det-cat-id',
        type=int,
        default=1,
        help='Category id for bounding box detection model')
    parser.add_argument(
        '--bbox-thr',
        type=float,
        default=0.3,
        help='Bounding box score threshold')
    parser.add_argument(
        '--kpt-thr', type=float, default=0.3, help='Keypoint score threshold')
    parser.add_argument(
        '--radius',
        type=int,
        default=4,
        help='Keypoint radius for visualization')
    parser.add_argument(
        '--thickness',
        type=int,
        default=1,
        help='Link thickness for visualization')

    assert has_mmdet, 'Please install mmdet to run the script.'

    args = parser.parse_args()

    assert args.det_config is not None
    assert args.det_checkpoint is not None

    # Initialise YARP
    yarp.Network.init()

    # Create a port and connect it to the realsense
    input_port = yarp.Port()
    input_port.open("/keypointsDetector/rgbImage:i")
    # yarp.Network.connect("/yarpOpenPose/image:o", input_port.getName())
    yarp.Network.connect("/depthCamera/rgbImage:o", input_port.getName())

    # Create numpy array to receive the image and the YARP image wrapped around it
    img_array = numpy.zeros((720, 1280, 3), dtype=numpy.uint8)
    yarp_image = yarp.ImageRgb()
    yarp_image.resize(1280, 720)
    yarp_image.setExternal(img_array, img_array.shape[1], img_array.shape[0])

    # Create output visualize port and connect it to the yarpview figure
    output_port = yarp.Port()
    output_port.open("/keypointsDetector/rgbImage:o")
    yarp.Network.connect(output_port.getName(), "/detectedKeypoints")
    
    # Create output keypoints port and connect it further
    keypoints_port = yarp.Port()
    keypoints_port.open("/keypointsDetector/keypoints")
    yarp.Network.connect(keypoints_port.getName(), "/keypoints_out") #TODO: specify for 2D->3D position script later on
    
    # Prepare MMPose detector
    det_model = init_detector(args.det_config, args.det_checkpoint, device=args.device)
    # build the pose model from a config file and a checkpoint file
    pose_model = init_pose_model(args.pose_config, args.pose_checkpoint, device=args.device)
    # specify forward solution using config file
    dataset = pose_model.cfg.data['test']['type']
    dataset_info = pose_model.cfg.data['test'].get('dataset_info', None)
    if dataset_info is None:
        warnings.warn(
            'Please set `dataset_info` in the config.'
            'Check https://github.com/open-mmlab/mmpose/pull/663 for details.',
            DeprecationWarning)
    else:
        dataset_info = DatasetInfo(dataset_info)
    cocoParts = ["Nose", "LEye", "REye", "LEar", "REar", "LShoulder", "RShoulder", "LElbow", "RElbow", "LWrist", "RWrist", "LHip", "RHip", "LKnee", "RKnee", "LAnkle", "RAnkle", "LBigToe", "LSmallToe", "LHeel", "RBigToe", "RSmallToe", "RHeel", "RJaw_0", "RJaw_1", "RJaw_2", "RJaw_3", "RJaw_4", "RJaw_5", "Chin_0", "Chin_1", "Chin_2", "Chin_3", "Chin_4", "LJaw_5", "LJaw_4", "LJaw_3", "LJaw_2", "LJaw_1", "LJaw_0", "Forehead_0", "Forehead_1", "Forehead_2", "Forehead_3", "Forehead_4", "Forehead_5", "Forehead_6", "Forehead_7", "Forehead_8", "Forehead_9", "FNose_0", "FNose_1", "FNose_2", "FNose_3", "FNose_4", "FNose_5", "FNose_6", "FNose_7", "FNose_8", "FREye_0", "FREye_1", "FREye_2", "FREye_3", "FREye_4", "FREye_5", "FLEye_0", "FLEye_1", "FLEye_2", "FLEye_3", "FLEye_4", "FLEye_5", "Lips_0", "Lips_1", "Lips_2", "Lips_3", "Lips_4", "Lips_5", "Lips_6", "Lips_7", "Lips_8", "Lips_9", "Lips_10", "Lips_11", "Mouth_0", "Mouth_1", "Mouth_2", "Mouth_3", "Mouth_4", "Mouth_5", "Mouth_6", "Mouth_7", "LHWrist", "LThumbCmc", "LThumbMcp", "LThumbIp", "LThumbTip", "LIndexFingerMcp", "LIndexFingerPip", "LIndexFingerDip", "LIndexFingerTip", "LMiddleFingerMcp", "LMiddleFingerPip", "LMiddleFingerDip", "LMiddleFingerTip", "LRingFingerMcp", "LRingFingerPip", "LRingFingerDip", "LRingFingerTip", "LPinkyMcp", "LPinkyDip", "LPinkyPip", "LPinkyTip", "RHWrist", "RThumbCmc", "RThumbMcp", "RThumbIp", "RThumbTip", "RIndexFingerMcp", "RIndexFingerPip", "RIndexFingerDip", "RIndexFingerTip", "RMiddleFingerMcp", "RMiddleFingerPip", "RMiddleFingerDip", "RMiddleFingerTip", "RRingFingerMcp", "RRingFingerPip", "RRingFingerDip", "RRingFingerTip", "RPinkyMcp", "RPinkyDip", "RPinkyPip", "RPinkyTip"]
    # Infinite loop for reading image from port and process it with detector
    while True:
        # Read the data from the port into the image
        input_port.read(yarp_image)
        # Process read image with detector
        mmdet_results = inference_detector(det_model, img_array)
        person_results = process_mmdet_results(mmdet_results, 1)

        pose_results, returned_outputs = inference_top_down_pose_model(
            pose_model, img_array, person_results, bbox_thr=0.3,
            dataset=dataset, dataset_info=dataset_info, return_heatmap=False,
            outputs=None
        )
        
        # Process output results
        if args.show:
            img_out = vis_pose_result(
                pose_model, img_array, pose_results, dataset=dataset,
                dataset_info=dataset_info, kpt_score_thr=0.3, radius=4,
                thickness=1, show=False, out_file=None
            )
            img_array[:] = img_out
            output_port.write(yarp_image)

        #TODO: Process or (and?) feedforward detected keypoints
        # first shot ....
        # best_score = 0
        # best_i = None
        # for i in range(len(pose_results)):
          #  score = pose_results[i]['bbox'][-1]
           # if score > best_score:
            #    best_score = score
             #   best_i = i
        #if best_i is None:
         #   continue
        for i in range(len(pose_results)):
            best_kpts = pose_results[i]['keypoints'].reshape(-1)
            x_kpts = best_kpts[0::3]
            y_kpts = best_kpts[1::3]
            score_kpts = best_kpts[2::3]
            NUM_KPTS = 133
            bout = yarp.Bottle()
            kpts = bout.addList()
            for i in range(NUM_KPTS):
                kpt = kpts.addList()
                kpt.addString(cocoParts[i])
                kpt.addDouble(float(x_kpts[i]))
                kpt.addDouble(float(y_kpts[i]))
                kpt.addDouble(float(score_kpts[i]))
            keypoints_port.write(bout)

    # Cleanup
    input_port.close()
    output_port.close()

if __name__ == '__main__':
    main()
