#!/usr/bin/python3
import sys
import numpy as np
import skimage.feature
import yarp
import matplotlib.pyplot as plt
import cv2
import time
import os
from sys import platform
from argparse import ArgumentParser
# import mediapipe as mp
import warnings
from copy import deepcopy


def main():


    # Import Openpose (Windows/Ubuntu/OSX)
    dir_path = os.path.dirname(os.path.realpath(__file__))
    try:
        # Windows Import
        if platform == "win32":
            # Change these variables to point to the correct folder (Release/x64 etc.)
            sys.path.append(dir_path + '/../../python/openpose/Release');
            os.environ['PATH']  = os.environ['PATH'] + ';' + dir_path + '/../../x64/Release;' +  dir_path + '/../../bin;'
            import pyopenpose as op
        else:
            # Change these variables to point to the correct folder (Release/x64 etc.)
            # If you run `make install` (default path is `/usr/local/python` for Ubuntu), you can also access the OpenPose/python module from there. This will install OpenPose and the python library at your desired installation path. Ensure that this is in your python path in order to use it.
            sys.path.append('/usr/local/python')
            from openpose import pyopenpose as op
    except ImportError as e:
        print('Error: OpenPose library could not be found. Did you enable `BUILD_PYTHON` in CMake and have this Python script in the right folder?')
        raise e
    # Custom Params (refer to include/openpose/flags.hpp for more parameters)
    params = dict()
    params["model_folder"] = "/home/icub/Documents/detectors/openpose/models/"
    params["face"] = False
    params["hand"] = True


    # Construct it from system arguments
    # op.init_argv(args[1])
    # oppython = op.OpenposePython()

    # Starting OpenPose
    opWrapper = op.WrapperPython()
    opWrapper.configure(params)
    opWrapper.start()

    # Process Image
    datum = op.Datum()
    # Initialise YARP
    yarp.Network.init()

    # Create a port and connect it to the realsense
    input_port = yarp.Port()
    input_port.open("/keypointsDetector/rgbImage:i")
    # yarp.Network.connect("/yarpOpenPose/image:o", input_port.getName())
    yarp.Network.connect("/depthCamera/rgbImage:o", input_port.getName())

    # Create numpy array to receive the image and the YARP image wrapped around it
    img_array = np.zeros((720, 1280, 3), dtype=np.uint8)
    yarp_image = yarp.ImageRgb()
    yarp_image.resize(1280, 720)
    yarp_image.setExternal(img_array, img_array.shape[1], img_array.shape[0])

    # Create output visualize port and connect it to the yarpview figure
    output_port = yarp.Port()
    output_port.open("/keypointsDetector/rgbImage:o")
    yarp.Network.connect(output_port.getName(), "/detectedKeypoints")
    
    # Create output keypoints port and connect it further
    keypoints_port = yarp.Port()
    keypoints_port.open("/keypointsDetector/keypoints")
    yarp.Network.connect(keypoints_port.getName(), "/keypoints_out") #TODO: specify for 2D->3D position script later on
    
    cocoParts = ["Nose", "LEye", "REye", "LEar", "REar", "LShoulder", "RShoulder", "LElbow", "RElbow", "LWrist", "RWrist", "LHip", "RHip", "LKnee", "RKnee", "LAnkle", "RAnkle", "LBigToe", "LSmallToe", "LHeel", "RBigToe", "RSmallToe", "RHeel", "RJaw_0", "RJaw_1", "RJaw_2", "RJaw_3", "RJaw_4", "RJaw_5", "Chin_0", "Chin_1", "Chin_2", "Chin_3", "Chin_4", "LJaw_5", "LJaw_4", "LJaw_3", "LJaw_2", "LJaw_1", "LJaw_0", "Forehead_0", "Forehead_1", "Forehead_2", "Forehead_3", "Forehead_4", "Forehead_5", "Forehead_6", "Forehead_7", "Forehead_8", "Forehead_9", "FNose_0", "FNose_1", "FNose_2", "FNose_3", "FNose_4", "FNose_5", "FNose_6", "FNose_7", "FNose_8", "FREye_0", "FREye_1", "FREye_2", "FREye_3", "FREye_4", "FREye_5", "FLEye_0", "FLEye_1", "FLEye_2", "FLEye_3", "FLEye_4", "FLEye_5", "Lips_0", "Lips_1", "Lips_2", "Lips_3", "Lips_4", "Lips_5", "Lips_6", "Lips_7", "Lips_8", "Lips_9", "Lips_10", "Lips_11", "Mouth_0", "Mouth_1", "Mouth_2", "Mouth_3", "Mouth_4", "Mouth_5", "Mouth_6", "Mouth_7", "LHWrist", "LThumbCmc", "LThumbMcp", "LThumbIp", "LThumbTip", "LIndexFingerMcp", "LIndexFingerPip", "LIndexFingerDip", "LIndexFingerTip", "LMiddleFingerMcp", "LMiddleFingerPip", "LMiddleFingerDip", "LMiddleFingerTip", "LRingFingerMcp", "LRingFingerPip", "LRingFingerDip", "LRingFingerTip", "LPinkyMcp", "LPinkyDip", "LPinkyPip", "LPinkyTip", "RHWrist", "RThumbCmc", "RThumbMcp", "RThumbIp", "RThumbTip", "RIndexFingerMcp", "RIndexFingerPip", "RIndexFingerDip", "RIndexFingerTip", "RMiddleFingerMcp", "RMiddleFingerPip", "RMiddleFingerDip", "RMiddleFingerTip", "RRingFingerMcp", "RRingFingerPip", "RRingFingerDip", "RRingFingerTip", "RPinkyMcp", "RPinkyDip", "RPinkyPip", "RPinkyTip"]
    
    bodyParts = ["Nose", "Neck", "RShoulder", "RElbow", "RWrist", "LShoulder", "LElbow", "LWrist", "Hip", "RHip", "RKnee", "RAnkle", "LHip", "LKnee", "LAnkle", "REye", "LEye", "REar", "LEar", "LBigToe", "LSmallToe", "LHeel", "RBigToe", "RSmallToe", "RHeel"]
    leftHandParts = ["LHWrist", "LThumbCmc", "LThumbMcp", "LThumbIp", "LThumbTip", "LIndexFingerMcp", "LIndexFingerPip", "LIndexFingerDip", "LIndexFingerTip", "LMiddleFingerMcp", "LMiddleFingerPip", "LMiddleFingerDip", "LMiddleFingerTip", "LRingFingerMcp", "LRingFingerPip", "LRingFingerDip", "LRingFingerTip", "LPinkyMcp", "LPinkyDip", "LPinkyPip", "LPinkyTip"]
    rightHandParts = ["RHWrist", "RThumbCmc", "RThumbMcp", "RThumbIp", "RThumbTip", "RIndexFingerMcp", "RIndexFingerPip", "RIndexFingerDip", "RIndexFingerTip", "RMiddleFingerMcp", "RMiddleFingerPip", "RMiddleFingerDip", "RMiddleFingerTip", "RRingFingerMcp", "RRingFingerPip", "RRingFingerDip", "RRingFingerTip", "RPinkyMcp", "RPinkyDip", "RPinkyPip", "RPinkyTip"]
    
    # Infinite loop for reading image from port and process it with detector
    while True:
        # Read the data from the port into the image
        input_port.read(yarp_image)
        # imageToProcess = deepcopy(image)
        datum = op.Datum()
        image_array = np.zeros_like(img_array)
        image_array[:] = img_array
        image = cv2.cvtColor(image_array, cv2.COLOR_BGR2RGB)
        datum.cvInputData = image
        opWrapper.emplaceAndPop(op.VectorDatum([datum]))
        # print(datum.poseKeypoints.shape)
        # print(datum.faceKeypoints.shape)
        # print(datum.handKeypoints[0].shape)
        # print(datum.handKeypoints[1].shape)
        
        img_array[:] = cv2.cvtColor(datum.cvOutputData, cv2.COLOR_RGB2BGR)
        
        
        if datum.poseKeypoints is not None:
            for i in range(datum.poseKeypoints.shape[0]):
                body_kpts = datum.poseKeypoints[i]
                x_kpts_body = body_kpts[:, 0]
                y_kpts_body = body_kpts[:, 1]
                score_kpts_body = body_kpts[:, 2]

                left_hand_kpts = datum.handKeypoints[0][i]
                x_kpts_lhand = left_hand_kpts[:, 0]
                y_kpts_lhand = left_hand_kpts[:, 1]
                score_kpts_lhand = left_hand_kpts[:, 2]
                
                right_hand_kpts = datum.handKeypoints[1][i]
                x_kpts_rhand = right_hand_kpts[:, 0]
                y_kpts_rhand = right_hand_kpts[:, 1]
                score_kpts_rhand = right_hand_kpts[:, 2]
                
                bout = yarp.Bottle()
                kpts = bout.addList()
                
                for j in range(25):
                    kpt = kpts.addList()
                    kpt.addString(bodyParts[j])
                    kpt.addDouble(float(x_kpts_body[j]))
                    kpt.addDouble(float(y_kpts_body[j]))
                    kpt.addDouble(float(score_kpts_body[j]))
                for j in range(21):
                    kpt = kpts.addList()
                    kpt.addString(leftHandParts[j])
                    kpt.addDouble(float(x_kpts_lhand[j]))
                    kpt.addDouble(float(y_kpts_lhand[j]))
                    kpt.addDouble(float(score_kpts_lhand[j]))
                for j in range(21):
                    kpt = kpts.addList()
                    kpt.addString(rightHandParts[j])
                    kpt.addDouble(float(x_kpts_rhand[j]))
                    kpt.addDouble(float(y_kpts_rhand[j]))
                    kpt.addDouble(float(score_kpts_rhand[j]))
                keypoints_port.write(bout)
        output_port.write(yarp_image)
        

    # Cleanup
    input_port.close()
    output_port.close()

if __name__ == '__main__':
    main()
